function [ mask ] = createMasks( im, params )
% Create masks interactively
%   by drawing foreground and background labels into the input image.
%
%   (c) 2015 Robert Bensch, Computer Science Department, University of Freiburg, Germany
%
  
  scrsz = get(0,'MonitorPosition');
  
  %
  % initialize histograms
  %
  nBins = params.nBins;
  bin_step = 1/nBins;
  bin_range = [0:bin_step:1];
  hist_binCenters = bin_range(1:end-1) + bin_step/2;
  
  %
  % specify which type of labels to create
  %
  doLabel = true;
  doLabelFgr = true;
  doLabelBgr = true;
  
  %
  % initialize foreground and background mask
  %
  mask = zeros(size(im));
  
  %
  % initialize figure
  %
  figure(1);
  im_overlay = cat(3, im+0.1*double(mask==1), im+0.1*double(mask==2), im);
  im_overlay = im_overlay / max(im_overlay(:));
  h = image( im_overlay);
  set(1,'Position',scrsz(1,:));
  axis off; axis image;
  title('Add foreground label (draw polygon and double-click to finish)');
  
  while (doLabel)
  
  %
  % add foreground labels
  %
  while (doLabelFgr)
    figure(1);
    im_overlay = cat(3, im+0.1*double(mask==1), im+0.1*double(mask==2), im);
    im_overlay = im_overlay / max(im_overlay(:));
    image(im_overlay);
    axis off; axis image;
    title('Add foreground label (draw polygon and double-click to finish)');
    
    h = impoly;
    %h = imfreehand;
    position = wait(h);
    freehand_mask = createMask(h);
    update_mask = (mask==2) | freehand_mask;
    mask(update_mask) = 2;
    im_overlay = cat(3, im+0.1*double(mask==1), im+0.1*double(mask==2), im);
    im_overlay = im_overlay / max(im_overlay(:));
    image(im_overlay);
    axis off; axis image;
    
    % compute histograms
    [hist_fgr, ~] = hists( im(mask==2), hist_binCenters);
    [hist_bgr, ~] = hists( im(mask==1), hist_binCenters);
    if (sum(hist_fgr)>0)
      hist_fgr = hist_fgr./sum(hist_fgr);
    end
    if (sum(hist_bgr)>0)
      hist_bgr = hist_bgr./sum(hist_bgr);
    end
    % plot histograms
    figure(2);
    hArray = bar(hist_binCenters', [hist_fgr; hist_bgr]');
    title('Foreground/background histograms');
    grid on;
    set(hArray(1), 'FaceColor', 'green', 'EdgeColor', 'green');
    set(hArray(2), 'FaceColor', 'red', 'EdgeColor', 'red');
    ylim([0 1]);
    xlabel('Intensity (normalized)'); ylabel('Relative frequency');
    legend('Foreground histogram','Background histogram');
    % compute foreground/background likelihood and foreground probability
    [~, binWeights] = hists( im(:), hist_binCenters);
    fgrLh = sum(bsxfun(@times, binWeights, hist_fgr), 2);
    bgrLh = sum(bsxfun(@times, binWeights, hist_bgr), 2);
    fgrPr = fgrLh ./ (fgrLh + bgrLh);
    figure(3);
    subplot(2,2,1); imshow(reshape(fgrLh,size(im)),[0 1]); colormap('jet'); colorbar; title('Foreground likelihood');
    subplot(2,2,2); imshow(reshape(bgrLh,size(im)),[0 1]); colormap('jet'); colorbar; title('Background likelihood');
    subplot(2,2,[3,4]); imshow(reshape(fgrPr,size(im)),[0 1]); colormap('jet'); colorbar; title('Foreground probability');
        
    figure(1);
    title('Add another foreground label?');
    
    choice = questdlg(['Add another foreground label?'], 'Create foreground mask',...
                            'Yes','No','No');
    switch choice
        case 'Yes'
            doLabelFgr = true;
        case 'No'
            doLabelFgr = false;
    end   

  end
  
  %
  % add background labels
  %
  while (doLabelBgr)
    figure(1);
    im_overlay = cat(3, im+0.1*double(mask==1), im+0.1*double(mask==2), im);
    im_overlay = im_overlay / max(im_overlay(:));
    image(im_overlay);
    axis off; axis image;
    title('Add background label (draw polygon and double-click to finish)');
    
    h = impoly;
    %h = imfreehand;
    position = wait(h);
    freehand_mask = createMask(h);
    update_mask = (mask==1) | freehand_mask;
    mask(update_mask) = 1;
    im_overlay = cat(3, im+0.1*double(mask==1), im+0.1*double(mask==2), im);
    im_overlay = im_overlay / max(im_overlay(:));
    image(im_overlay);
    axis off; axis image;
    
    % compute histograms
    [hist_fgr, ~] = hists( im(mask==2), hist_binCenters);
    [hist_bgr, ~] = hists( im(mask==1), hist_binCenters);
    if (sum(hist_fgr)>0)
      hist_fgr = hist_fgr./sum(hist_fgr);
    end
    if (sum(hist_bgr)>0)
      hist_bgr = hist_bgr./sum(hist_bgr);
    end
    % plot histograms
    figure(2);
    hArray = bar(hist_binCenters', [hist_fgr; hist_bgr]');
    title('Foreground/background histograms');
    grid on;
    set(hArray(1), 'FaceColor', 'green', 'EdgeColor', 'green');
    set(hArray(2), 'FaceColor', 'red', 'EdgeColor', 'red');
    ylim([0 1]);
    xlabel('Intensity (normalized)'); ylabel('Relative frequency');
    legend('Foreground histogram','Background histogram');
    % compute foreground/background likelihood and foreground probability
    [~, binWeights] = hists( im(:), hist_binCenters);
    fgrLh = sum(bsxfun(@times, binWeights, hist_fgr), 2);
    bgrLh = sum(bsxfun(@times, binWeights, hist_bgr), 2);
    fgrPr = fgrLh ./ (fgrLh + bgrLh);
    figure(3);
    subplot(2,2,1); imshow(reshape(fgrLh,size(im)),[0 1]); colormap('jet'); colorbar; title('Foreground likelihood');
    subplot(2,2,2); imshow(reshape(bgrLh,size(im)),[0 1]); colormap('jet'); colorbar; title('Background likelihood');
    subplot(2,2,[3,4]); imshow(reshape(fgrPr,size(im)),[0 1]); colormap('jet'); colorbar; title('Foreground probability');
    
    figure(1);
    title('Add another background label?');
    
    choice = questdlg(['Add another background label?'], 'Create background mask',...
                            'Yes','No','No');
    switch choice
        case 'Yes'
            doLabelBgr = true;
        case 'No'
            doLabelBgr = false;
    end

  end
  
   choice = questdlg(['Add more labels?'], 'Create mask',...
                            'Yes','No','No');
    switch choice
        case 'Yes'
            doLabel = true;
            doLabelFgr = true;
            doLabelBgr = true;
        case 'No'
            doLabel = false;
    end   
  
  end
  
  close(1);
  close(2);
  close(3);
  
end

