function [predLabels, fileLst] = segmentCells( fileStr, params )
% Runs cell segmentation
%
%   fileStr: input files string (e.g. 'test/t*.tif')
%   params: parameter structure
%       binaryTermType
%         'symmetric': symmetric boundary costs (first case in Equ. 2)
%         'asymmetric-darkToBright': asymmetric boundary costs with
%                                    dark-to-bright preference (Equ. 2)
%         'asymmetric-brightToDark': asymmetric boundary costs with
%                                    bright-to-dark preference
%
%   output
%     predLabels: image stack (nCols, nRows, nFrames) containing the
%       segmented frames (segment labels from 1...K, K maximum label, background label = 0, datatyp uint16)
%     fileLst: cell array containing processed files
%
%   (c) 2015 Robert Bensch, Computer Science Department, University of Freiburg, Germany
%

addpath('maxflow/');

%
% Parameters
%
model = params.model;

% Pre-processing
doPreSmoothing = params.doPreSmoothing;
preSmoothSigma = params.preSmoothSigma;

% Binary term
nNeigh = params.nNeigh;
binaryTermType = params.binaryTermType;
sigma = params.sigma;

% Lambda
lambda = params.lambda;

% Post-processing
useFillHoles = params.useFillHoles;
minSegmAreaPx = params.minSegmAreaPx;

% Tracking / Segmentation propagation
useFgrPropagation = params.useFgrPropagation;
erosion_size = params.erosion_size;
useNonMergingConstr = params.useNonMergingConstr;

debugMode = params.debugMode;


% Precompute filter kernels 
hsize = ceil(2*(3*preSmoothSigma));
if (mod(hsize,2)==0)
    hsize = hsize + 1;
end
hsize_preSmoothSigma = hsize;
kernel_preSmoothing = fspecial('gaussian', hsize_preSmoothSigma, preSmoothSigma);

% Precompute structuring elements
if (useFgrPropagation)
    seErosion = strel('disk',erosion_size);  
end

% Get input files
d = dir( fileStr);
nFrames = length(d);
% Get path
idx = strfind(fileStr, '/');
path = '';
if (~isempty(idx))
    path = fileStr(1:idx(end));
end

%
% Run segmentation
%
tstart_total = tic;

fprintf('processing %d file(s) in total.\n', nFrames);

fileLst = [];
E = [];
dist_pq = [];

if (debugMode > 0)
  figHdl = figure;
end

for frameIdx=1:nFrames
    
  tstart = tic;    
    
  fn = d(frameIdx).name;
  
  fprintf('\tprocessing %s\n', fn);
  
  % (nCols, nRows)
  im = im2double(imread([path fn])');
  if (debugMode > 0)
    im_raw = im;
  end
    
  if (model.params.doPreSmoothing)
    im_regionPrior = imfilter( im, model.params.kernel_preSmoothing, 'symmetric');    
  else
    im_regionPrior = im;
  end
  if (model.params.doBackgroundSubstr)
    im_bgr = imfilter( im_regionPrior, model.params.kernel_bgrSubstr, 'symmetric');  
    im_regionPrior = 0.5 * ((im_regionPrior - im_bgr) + 1.0);    % > [0,2]    
    im = im_regionPrior;    
  end

  if (doPreSmoothing)
    im = imfilter( im, kernel_preSmoothing, 'symmetric');    
  end
  
  imsize = size(im);
  
  if (frameIdx==1)
    % Initialize predLabels stack (nCols, nRows, nLevs)
    predLabels = uint16( zeros(imsize(1), imsize(2), nFrames));
  end
  
  % Create graph edges
  if (frameIdx==1)
    if (nNeigh==4)
      E = edges4connected(imsize(1), imsize(2));
    elseif (nNeigh==8)
      E = edges8connected(imsize(1), imsize(2));
    end

    % Encoding of edge directions
    %   left, right, up, down : -1,1,-2,2
    %   diagonal left up/down, right up/down: -3, 3, -4, 4
    dirE = (E(:,2) - E(:,1));
    dirE(dirE==imsize(1)) = 2;
    dirE(dirE==-imsize(1)) = -2;
    if (nNeigh==8)
        % diagonal edges
        dirE(dirE==-(imsize(1)-1)) = -3;
        dirE(dirE==-(imsize(1)+1)) = -4;
        dirE(dirE==(imsize(1)-1)) = 3;
        dirE(dirE==(imsize(1)+1)) = 4;
    end
    
    dist_pq = ones(size(E,1),1);
    if (nNeigh==8)
        dist_pq(abs(dirE)>2) = sqrt(2);
    end
    
  end
  
  %
  % Compute image intensity gradient for each edge
  %
  g = im( E(:,2)) - im(E(:,1));
   
  %
  % Compute edge costs
  %
  if (strcmp(binaryTermType,'symmetric'))
    % Symmetric binary costs
    weights = exp( - g.^2 ./ (2*sigma^2));
  elseif (strcmp(binaryTermType,'asymmetric-darkToBright'))       
    % Asymmetric binary costs
    weights = ones(size(g));
    weights(g>0) = exp( - g(g>0).^2 ./ (2*sigma^2));
  elseif (strcmp(binaryTermType,'asymmetric-brightToDark'))
    % Asymmetric binary costs
    weights = ones(size(g));
    weights(g<0) = exp( - g(g<0).^2 ./ (2*sigma^2));
  else
     error('Not implemented!');
  end
  % Normalize by neighbor distances
  weights = weights .* (1./dist_pq);
     
  %
  % Graph construction
  %
  A = sparse(E(:,1),E(:,2),weights,prod(imsize),prod(imsize),nNeigh*prod(imsize));   
  tt = zeros( imsize);
  T = sparse( prod(imsize), 2);

  %
  % Segmentation propagation
  %
  if (frameIdx==1)
    %
  else
    if (useFgrPropagation)
        % Propagate previous segmentation mask using hard foreground
        % constraints for the next frame
        tt(L2_eroded) = 1;
    end
    if (useNonMergingConstr)
        % Propagate boundaries of previous segments "support regions" using
        % hard background constraints for the next frame
        D_L2_prev = bwdist(L2_prev>0);
        DL_L2_prev = watershed(D_L2_prev);
        RL = DL_L2_prev == 0;
        tt(RL) = 2;
    end
  end
   
  % hard background constraint: image border is background
  tt(:,1:2) = 2;
  tt(1:2,:) = 2;
  tt(:,end-1:end) = 2;
  tt(end-1:end,:) = 2;
  
  % get hard foreground and background pixels
  tt_1 = (tt==1);
  tt_1 = tt_1(:);
  tt_2 = (tt==2);
  tt_2 = tt_2(:);   

  %
  % Compute data costs / regional penalties
  %
  [~, binWeights] = hists( im_regionPrior(:), model.hist_binCenters);
  fgrLh = sum(bsxfun(@times, binWeights, model.hist_fgr), 2);  %fgrLh = sum(gmultiply( binWeights, model.fgrPr), 2);
  bgrLh = sum(bsxfun(@times, binWeights, model.hist_bgr), 2);  %bgrLh = sum(gmultiply( binWeights, model.bgrPr), 2);
    
  % probabilities:
  fgrPr = fgrLh ./ (fgrLh + bgrLh);    
  fgrRp = 1-fgrPr;
  bgrRp = fgrPr;
  
%   % negative log-likelihoods:
%   fgrRp = -log(fgrLh);
%   bgrRp = -log(bgrLh);
  
  T(:,1) = bgrRp;     % s-link costs
  T(:,2) = fgrRp;     % t-link costs


  %
  % Weighting region term by lambda
  % 
  T = lambda .* T;
          
  %
  % Set hard foreground and background constraints
  %

  % compute constant K
  K = 1 + nNeigh*max(weights);

  % foreground
  T(tt_1, 1) = K;
  T(tt_1, 2) = 0;
  % background
  T(tt_2, 1) = 0;
  T(tt_2, 2) = K;
   
  %
  % Run maxflow
  %
  [~,labels] = maxflow(A,T);
  %L = double(1-reshape(labels, imsize));
  L = (1-reshape(labels, imsize)) > 0;
    
  %
  % Fill holes
  %
  if (useFillHoles==1)
    L = imfill(L,'holes');
  end
  CC = bwconncomp(L);
  
  %
  % Remove small segments
  %
  if (minSegmAreaPx > 0)
    numPixels = cellfun(@numel,CC.PixelIdxList);
    CC2 = CC;
    CC2.PixelIdxList = CC2.PixelIdxList(numPixels>minSegmAreaPx);
    CC2.NumObjects = length(CC2.PixelIdxList);
    CC = CC2;
  end 
  
  L2 = labelmatrix( CC);
        
  if (useFgrPropagation)
      % Erode segmentation mask
      L2_eroded = imerode(L2>0,seErosion);
  end
    
  %
  % Write segmentation mask (labels, uint16)
  %
  predLabels(:,:,frameIdx) = uint16(L2);
  fileLst{frameIdx} = fn;
  
%   if (frameIdx < 6)
%   close all
%   figure; imshow(im',[]);
%   figure; imshow(L2',[]);
%   pause(0.5);  
%   end
  
  if (debugMode > 0)
    im_overlay = cat(3, im_raw'+0.4*double(L2'>0), im_raw', im_raw');
    im_overlay = im_overlay / max(im_overlay(:));
    figure(figHdl); imshow(im_overlay);
    title('Cell segmentation (mask in red)');
    xlabel(strrep(fn,'_',' '));
    pause(0.5);
  end

  % Store segmentation mask to be used in the next run
  %
  L2_prev = L2;
  
  tend = toc(tstart);
  fprintf('\ttime (sec): %5.2f\n', tend);
  
end

tend_total = toc(tstart_total);
fprintf('time total (sec): %5.2f\n', tend_total);

end

